#include "cudadevice.h"

CudaDevice::CudaDevice(int __id) : isInit(false), isContext(false), _id(__id), _cMaj(0), _cMin(0)
{
  posDevPtr = 0;

  CUresult  result;
  char      buffer[CUDA_CHAR_BUFFER_SIZE];
  memset(attribute, 0, sizeof(attribute));

  if((result = cuDeviceGet(&device, _id )) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("Error: %1 in method cuDeviceGet").arg(errorString(result));
    return;
  }

  if((result = cuDeviceGetName(buffer, CUDA_CHAR_BUFFER_SIZE, device)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("Error: %1 in method cuDeviceGetName").arg(errorString(result));
    return;
  }

  if((result = cuDeviceComputeCapability(&_cMaj, &_cMin, device)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("Error: %1 in method cuDeviceComputeCapability").arg(errorString(result));
    return;
  }

  if((result = cuDeviceGetProperties(&props, device)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("Error: %1 in method cuDeviceGetProperties").arg(errorString(result));
    return;
  }

  if((result = cuDeviceTotalMem(&_totalMem, device)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("Error: %1 in method cuDeviceTotalMem").arg(errorString(result));
    return;
  }

  for (int i = CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK ; i <= CU_DEVICE_ATTRIBUTE_TCC_DRIVER ; i++){
    if((result = cuDeviceGetAttribute(&attribute[i], (CUdevice_attribute)i, device)) != CUDA_SUCCESS){
      qDebug() << QObject::trUtf8("%1 in method cuDeviceGetAttribute for attribute %2").arg(errorString(result)).arg(attributeString(i));
      return;
    }
  }
  _name  = buffer;
  isInit = true;
}

bool CudaDevice::createContext(){
  CUresult result;
  if(!isInit)return false;
  if(isContext){
    qDebug() << QObject::trUtf8("Warning: CUDA context, already created !");
    return true;
  }
  if((result = cuCtxCreate(&context, CU_CTX_SCHED_AUTO, device)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuCtxCreate").arg(errorString(result));
    return false;
  }
  isContext = true;
  if((result = cuMemGetInfo(&_infoMemFree, &_infoMemTotal)) == CUDA_SUCCESS) return true;
  qDebug() << QObject::trUtf8("%1 in method cuMemGetInfo").arg(errorString(result));
  detachContext();
  return false;
}

bool CudaDevice::detachContext(){
  if(!isContext){
    qDebug() << QObject::trUtf8("Warning: Cant't detach CUDA context, already detached !");
    return true;
  }
  CUresult result;
//if((result = cuCtxDetach(context)) != CUDA_SUCCESS){
  if((result = cuCtxDestroy(context)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuCtxDestroy").arg(errorString(result));
    return false;
  }
  isContext = false;
  return true;
}

bool CudaDevice::loadModule(QString fileName){
  CUresult result;
  if((result = cuModuleLoad(&module, fileName.toUtf8().data())) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuModuleLoad").arg(errorString(result));
    return false;
  }
  return true;
}

bool CudaDevice::freeModule(){
  CUresult result;
  if((result = cuModuleUnload(module)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuModuleUnload").arg(errorString(result));
    return false;
  }
  return true;
}

bool CudaDevice::memAlloc(CUdeviceptr * ptr, size_t memSize){
  CUresult result;
  if((result = cuMemAlloc(ptr, memSize)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuMemAlloc").arg(errorString(result));
    return false;
  }
  return true;
}

bool CudaDevice::memFree(CUdeviceptr ptr){
  CUresult result;
  if((result = cuMemFree(ptr)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuMemFree").arg(errorString(result));
    return false;
  }
  return true;
}

bool CudaDevice::memcpyHtoD(CUdeviceptr ptr, void * src, size_t memSize){
  CUresult result;
  if((result = cuMemcpyHtoD(ptr, src, memSize)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuMemcpyHtoD").arg(errorString(result));
    return false;
  }
  return true;
}

bool CudaDevice::memcpyDtoH(void * out, CUdeviceptr ptr, size_t memSize){
  CUresult result;
  if((result = cuMemcpyDtoH(out, ptr, memSize)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuMemcpyDtoH").arg(errorString(result));
    return false;
  }
  return true;
}

bool CudaDevice::moduleGetFunction(const char * funName, CUfunction * func){
  CUresult result;
  if((result = cuModuleGetFunction(func, module, funName)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuModuleGetFunction").arg(errorString(result));
    return false;
  }
  return true;
}

bool CudaDevice::paramSetSize(CUfunction func, int offset){
  CUresult result;
  if((result = cuParamSetSize(func, offset)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuParamSetSize").arg(errorString(result));
    return false;
  }
  return true;
}

bool CudaDevice::funcSetBlockShape(CUfunction func, int x, int y, int z){
  CUresult result;
  if((result = cuFuncSetBlockShape(func, x, y, z)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuFuncSetBlockShape").arg(errorString(result));
    return false;
  }
  return true;
}

bool CudaDevice::funcSetSharedSize(CUfunction func, unsigned int numBytes){
  CUresult result;
  if((result = cuFuncSetSharedSize(func, numBytes)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuFuncSetSharedSize").arg(errorString(result));
    return false;
  }
  return true;
}

bool CudaDevice::paramSetI(CUfunction func, int * offset, size_t addOffset, int val, int * storeOffset){
  CUresult result = cuParamSeti(func, (*offset), val);
  if(result != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuParamSeti").arg(errorString(result));
    return false;
  }
  if(storeOffset != 0)(*storeOffset) = (*offset);
  (*offset) += addOffset;
  return true;
}

bool CudaDevice::paramSetUL(CUfunction func, int * offset, size_t addOffset, unsigned long val, int * storeOffset){
  CUresult result = cuParamSetv(func, (*offset), (void*)(&val), sizeof(unsigned long));
  if(result != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuParamSetv(unsigned long)").arg(errorString(result));
    return false;
  }
  if(storeOffset != 0)(*storeOffset) = (*offset);
  (*offset) += addOffset;
  return true;
}

bool CudaDevice::paramSetF(CUfunction func, int * offset, size_t addOffset, float val, int * storeOffset){
  CUresult result = cuParamSetf(func, (*offset), val);
  if(result != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuParamSetf").arg(errorString(result));
    return false;
  }
  if(storeOffset != 0)(*storeOffset) = (*offset);
  (*offset) += addOffset;
  return true;
}

bool CudaDevice::paramSetD(CUfunction func, int * offset, size_t addOffset, double val, int * storeOffset){
  CUresult result = cuParamSetf(func, (*offset), val);
  if(result != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuParamSetf(double)").arg(errorString(result));
    return false;
  }
  if(storeOffset != 0)(*storeOffset) = (*offset);
  (*offset) += addOffset;
  return true;
}

bool CudaDevice::paramSetV(CUfunction func, int * offset, size_t addOffset, void * ptr, unsigned int numBytes, int * storeOffset){
  CUresult result = cuParamSetv(func, (*offset), ptr, numBytes);
  if(result != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method cuParamSetv").arg(errorString(result));
    return false;
  }
  if(storeOffset != 0)(*storeOffset) = (*offset);
  (*offset) += addOffset;
  return true;
}

bool CudaDevice::launchGrid(CUfunction func, bool useGrid, int width, int height){
  QString  usedFunc = QString ("cuLaunch");
  CUresult result;
  if(useGrid){
    usedFunc.append("Grid");
    result = cuLaunchGrid(func, width, height);
  } else result = cuLaunch(func);
  if(result != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("%1 in method %2").arg(errorString(result)).arg(usedFunc);
    return false;
  }
  return true;
}
/*
bool CudaDevice::allocPos(VecF * data, int count){
  if(posDevPtr != 0){
    qDebug() << QObject::trUtf8("Cellule positions already allocated");
    return false;
  }
  size_t memSize = count * 3 * sizeof(float);
  if(!memAlloc(&posDevPtr, memSize))return false;
  if(!memcpyHtoD(posDevPtr, (void*)data, memSize))return false;
  return true;
}

bool CudaDevice::freePos(){
  if(posDevPtr == 0){
    qDebug() << QObject::trUtf8("Can't free device memory, cellule positions not allocated");
    return false;
  }
  if(!memFree(posDevPtr))return false;
  posDevPtr = 0;
  return true;
}
*/
bool CudaDevice::prepareCellPos(ProjectF & prj, VecF * allPositions, VecF & _begin, int stepAlpha, int stepBeta, int stepGama)
{
  bool result = false;
  QString kernel = "prepareCellPos";

  int offset = 0;
  int offset14, offset15, offset16;
  int reg = REGS__prepareCellPos;
  int maxThredPerBlock = props.regsPerBlock / ((reg == 0) ? 1 : reg);
  int maxUsedThred = stepAlpha;
  int maxBlockShape;
  CUfunction  func;
  CUdeviceptr ptr;
  VecF  begin(_begin.x, _begin.y, _begin.z);
  size_t memSize = stepAlpha * stepBeta * stepGama * 3 * sizeof(float);

  if(maxThredPerBlock > props.maxThreadsPerBlock)maxThredPerBlock = props.maxThreadsPerBlock;
  if(maxUsedThred > maxThredPerBlock)maxBlockShape = maxThredPerBlock;
  else maxBlockShape = maxUsedThred;

  if(!moduleGetFunction(kernel.toUtf8().constData(), &func))return result;
  if(!memAlloc(&ptr, memSize))return result;

  __ALIGN(ptr);         if(!paramSetV(func, &offset, sizeof(ptr),   (void*)&ptr,  sizeof(ptr))) goto ERROR_PrepareCellPos;
  __ALIGN(stepAlpha);   if(!paramSetI(func, &offset, sizeof(int),   stepAlpha    ))             goto ERROR_PrepareCellPos;
  __ALIGN(stepBeta);    if(!paramSetI(func, &offset, sizeof(int),   stepBeta     ))             goto ERROR_PrepareCellPos;
  __ALIGN(stepGama);    if(!paramSetI(func, &offset, sizeof(int),   stepGama     ))             goto ERROR_PrepareCellPos;
  __ALIGN(prj.alpha.x); if(!paramSetF(func, &offset, sizeof(float), prj.alpha.x  ))             goto ERROR_PrepareCellPos;
  __ALIGN(prj.alpha.y); if(!paramSetF(func, &offset, sizeof(float), prj.alpha.y  ))             goto ERROR_PrepareCellPos;
  __ALIGN(prj.alpha.z); if(!paramSetF(func, &offset, sizeof(float), prj.alpha.z  ))             goto ERROR_PrepareCellPos;
  __ALIGN(prj.beta.x);  if(!paramSetF(func, &offset, sizeof(float), prj.beta.x   ))             goto ERROR_PrepareCellPos;
  __ALIGN(prj.beta.y);  if(!paramSetF(func, &offset, sizeof(float), prj.beta.y   ))             goto ERROR_PrepareCellPos;
  __ALIGN(prj.beta.z);  if(!paramSetF(func, &offset, sizeof(float), prj.beta.z   ))             goto ERROR_PrepareCellPos;
  __ALIGN(prj.gama.x);  if(!paramSetF(func, &offset, sizeof(float), prj.gama.x   ))             goto ERROR_PrepareCellPos;
  __ALIGN(prj.gama.y);  if(!paramSetF(func, &offset, sizeof(float), prj.gama.y   ))             goto ERROR_PrepareCellPos;
  __ALIGN(prj.gama.z);  if(!paramSetF(func, &offset, sizeof(float), prj.gama.z   ))             goto ERROR_PrepareCellPos;
  __ALIGN(begin.x);     if(!paramSetF(func, &offset, sizeof(float), begin.x, &offset14))        goto ERROR_PrepareCellPos;
  __ALIGN(begin.y);     if(!paramSetF(func, &offset, sizeof(float), begin.y, &offset15))        goto ERROR_PrepareCellPos;
  __ALIGN(begin.z);     if(!paramSetF(func, &offset, sizeof(float), begin.z, &offset16))        goto ERROR_PrepareCellPos;

  if(!paramSetSize(func, offset))goto ERROR_PrepareCellPos;
  if(!funcSetBlockShape(func, maxBlockShape, 1, 1))goto ERROR_PrepareCellPos;

  if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellPos;

  if(!memcpyDtoH((void*)(&allPositions[0]), ptr, memSize))goto ERROR_PrepareCellPos;

  if(prj.matrixType == BCC){
    begin.x = _begin.x + 0.5f*prj.alpha.x + 0.5f*prj.beta.x + 0.5f*prj.gama.x;
    begin.y = _begin.y + 0.5f*prj.alpha.y + 0.5f*prj.beta.y + 0.5f*prj.gama.y;
    begin.z = _begin.z + 0.5f*prj.alpha.z + 0.5f*prj.beta.z + 0.5f*prj.gama.z;
    __ALIGN(begin.x); if(!paramSetF(func, &offset14, sizeof(float), begin.x)) goto ERROR_PrepareCellPos;
    __ALIGN(begin.y); if(!paramSetF(func, &offset15, sizeof(float), begin.y)) goto ERROR_PrepareCellPos;
    __ALIGN(begin.z); if(!paramSetF(func, &offset16, sizeof(float), begin.z)) goto ERROR_PrepareCellPos;
    if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellPos;
    if(!memcpyDtoH((void*)(&allPositions[stepAlpha*stepBeta*stepGama]), ptr, memSize))goto ERROR_PrepareCellPos;
  } else if(prj.matrixType == FCC){
    begin.x = _begin.x + 0.5f*prj.alpha.x + 0.5f*prj.beta.x;
    begin.y = _begin.y + 0.5f*prj.alpha.y + 0.5f*prj.beta.y;
    begin.z = _begin.z + 0.5f*prj.alpha.z + 0.5f*prj.beta.z;
    __ALIGN(begin.x); if(!paramSetF(func, &offset14, sizeof(float), begin.x)) goto ERROR_PrepareCellPos;
    __ALIGN(begin.y); if(!paramSetF(func, &offset15, sizeof(float), begin.y)) goto ERROR_PrepareCellPos;
    __ALIGN(begin.z); if(!paramSetF(func, &offset16, sizeof(float), begin.z)) goto ERROR_PrepareCellPos;
    if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellPos;
    if(!memcpyDtoH((void*)(&allPositions[stepAlpha*stepBeta*stepGama]), ptr, memSize))goto ERROR_PrepareCellPos;

    begin.x = _begin.x + 0.5f*prj.alpha.x + 0.5f*prj.gama.x;
    begin.y = _begin.y + 0.5f*prj.alpha.y + 0.5f*prj.gama.y;
    begin.z = _begin.z + 0.5f*prj.alpha.z + 0.5f*prj.gama.z;
    __ALIGN(begin.x); if(!paramSetF(func, &offset14, sizeof(float), begin.x)) goto ERROR_PrepareCellPos;
    __ALIGN(begin.y); if(!paramSetF(func, &offset15, sizeof(float), begin.y)) goto ERROR_PrepareCellPos;
    __ALIGN(begin.z); if(!paramSetF(func, &offset16, sizeof(float), begin.z)) goto ERROR_PrepareCellPos;
    if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellPos;
    if(!memcpyDtoH((void*)(&allPositions[2*stepAlpha*stepBeta*stepGama]), ptr, memSize))goto ERROR_PrepareCellPos;

    begin.x = _begin.x + 0.5f*prj.beta.x + 0.5f*prj.gama.x;
    begin.y = _begin.y + 0.5f*prj.beta.y + 0.5f*prj.gama.y;
    begin.z = _begin.z + 0.5f*prj.beta.z + 0.5f*prj.gama.z;
    __ALIGN(begin.x); if(!paramSetF(func, &offset14, sizeof(float), begin.x)) goto ERROR_PrepareCellPos;
    __ALIGN(begin.y); if(!paramSetF(func, &offset15, sizeof(float), begin.y)) goto ERROR_PrepareCellPos;
    __ALIGN(begin.z); if(!paramSetF(func, &offset16, sizeof(float), begin.z)) goto ERROR_PrepareCellPos;
    if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellPos;
    if(!memcpyDtoH((void*)(&allPositions[3*stepAlpha*stepBeta*stepGama]), ptr, memSize))goto ERROR_PrepareCellPos;
  }
  result = true;
ERROR_PrepareCellPos:
  if(!memFree(ptr))result = false;
  return result;
}

bool CudaDevice::prepareCellMate(ProjectF & prj, CellMate * allMates, int stepAlpha, int stepBeta, int stepGama){
  bool result = false;
  QString kernel = "prepareCellMate";

  int matesOffset[MAX_MATES], cellOffsetFun, offset = 0;
  int mates[MAX_MATES], cellOffset = 0;
  int reg = REGS__prepareCellMate;
  int maxThredPerBlock = props.regsPerBlock / ((reg == 0) ? 1 : reg);
  int maxUsedThred = stepAlpha;
  int maxBlockShape;
  int BxG = stepBeta * stepGama;
  size_t memSize = stepAlpha * BxG * sizeof(CellMate);
  CUdeviceptr ptr;
  CUfunction  func;

  if(maxThredPerBlock > props.maxThreadsPerBlock)maxThredPerBlock = props.maxThreadsPerBlock;
  if(maxUsedThred > maxThredPerBlock)maxBlockShape = maxThredPerBlock;
  else maxBlockShape = maxUsedThred;

  if(!moduleGetFunction(kernel.toUtf8().data(), &func))return result;
  if(!memAlloc(&ptr, memSize))return result;

  __ALIGN(ptr);       if(!paramSetV(func, &offset, sizeof(ptr), (void*)&ptr, sizeof(ptr)))goto ERROR_PrepareCellMate;
  __ALIGN(stepAlpha); if(!paramSetI(func, &offset, sizeof(int), stepAlpha))               goto ERROR_PrepareCellMate;
  __ALIGN(BxG);       if(!paramSetI(func, &offset, sizeof(int), BxG))                     goto ERROR_PrepareCellMate;

  prepareMateOffset(prj.matrixType, 1, mates, stepAlpha, stepBeta, stepGama);
  for(int i = 0 ; i < MAX_MATES ; i++){
    __ALIGN(mates[i]);
    if(!paramSetI(func,&offset, sizeof(int), mates[i], &matesOffset[i]))goto ERROR_PrepareCellMate;
  }

  __ALIGN(cellOffset);
  if(!paramSetI(func, &offset, sizeof(int), cellOffset, &cellOffsetFun))goto ERROR_PrepareCellMate;

  if(!paramSetSize(     func, offset))goto ERROR_PrepareCellMate;
  if(!funcSetBlockShape(func, maxBlockShape, 1, 1))goto ERROR_PrepareCellMate;

  if(!paramSetI(func, &cellOffsetFun, 0, cellOffset))goto ERROR_PrepareCellMate;
  if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellMate;
  if(!memcpyDtoH((void*)(&allMates[cellOffset]), ptr, memSize))goto ERROR_PrepareCellMate;


  prepareMateOffset(prj.matrixType, 2, mates, stepAlpha, stepBeta, stepGama);
  for(int i = 0 ; i < MAX_MATES ; i++){
    __ALIGN(mates[i]);
    if(!paramSetI(func, &matesOffset[i], 0, mates[i]))goto ERROR_PrepareCellMate;
  }
  cellOffset = stepAlpha * stepBeta * stepGama;
  __ALIGN(cellOffset);
  if(!paramSetI(func, &cellOffsetFun, 0, cellOffset))goto ERROR_PrepareCellMate;
  if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellMate;
  if(!memcpyDtoH((void*)(&allMates[cellOffset]), ptr, memSize))goto ERROR_PrepareCellMate;


  if(prj.matrixType == FCC)
  {
    prepareMateOffset(prj.matrixType, 3, mates, stepAlpha, stepBeta, stepGama);
    for(int i = 0 ; i < MAX_MATES ; i++){
      __ALIGN(mates[i]);
      if(!paramSetI(func, &matesOffset[i], 0, mates[i]))goto ERROR_PrepareCellMate;
    }
    cellOffset = 2 * stepAlpha * stepBeta * stepGama;
    __ALIGN(cellOffset);
    if(!paramSetI(func, &cellOffsetFun, 0, cellOffset))goto ERROR_PrepareCellMate;
    if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellMate;
    if(!memcpyDtoH((void*)(&allMates[cellOffset]), ptr, memSize))goto ERROR_PrepareCellMate;

    prepareMateOffset(prj.matrixType, 4, mates, stepAlpha, stepBeta, stepGama);
    for(int i = 0 ; i < MAX_MATES ; i++){
      __ALIGN(mates[i]);
      if(!paramSetI(func, &matesOffset[i], 0, mates[i]))goto ERROR_PrepareCellMate;
    }
    cellOffset = 3 * stepAlpha * stepBeta * stepGama;
    __ALIGN(cellOffset);
    if(!paramSetI(func, &cellOffsetFun, 0, cellOffset))goto ERROR_PrepareCellMate;
    if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellMate;
    if(!memcpyDtoH((void*)(&allMates[cellOffset]), ptr, memSize))goto ERROR_PrepareCellMate;
  }

  result = true;
  ERROR_PrepareCellMate:
  if(!memFree(ptr))result = false;
  return result;
}

bool CudaDevice::prepareCellMask(ProjectF & prj, VecF * allPositions, ulong * allMask, int stepAlpha, int stepBeta, int stepGama){
  bool result = false;
  QString kernel = "prepareCellMask";

  int offset = 0, maskBaseOffset;
  int reg = REGS__prepareCellMask;
  int maxThredPerBlock = props.regsPerBlock / ((reg == 0) ? 1 : reg);
  int maxUsedThred = stepAlpha;
  int maxBlockShape;
  int BxG   = stepBeta * stepGama;
  int AxBxG = stepAlpha * stepBeta * stepGama;

  ulong maskBase = CELL_MASK_NULL;

  size_t memSizeMask = AxBxG * sizeof(ulong);
  size_t memSizePos  = AxBxG * sizeof(VecF);
  CUdeviceptr ptrPos;
  CUdeviceptr ptrMask;
  CUfunction  func;

  if(maxThredPerBlock > props.maxThreadsPerBlock)maxThredPerBlock = props.maxThreadsPerBlock;
  if(maxUsedThred > maxThredPerBlock)maxBlockShape = maxThredPerBlock;
  else maxBlockShape = maxUsedThred;

  CellMaskSetCrystalStruct( &maskBase, prj.matrixType);
  CellMaskSetLattice(       &maskBase, 1);

  if(!moduleGetFunction(kernel.toUtf8().data(), &func))return result;
  if(!memAlloc(&ptrPos,  memSizePos))return result;
  if(!memAlloc(&ptrMask, memSizeMask))return result;

  if(!memcpyHtoD(ptrPos,  (void*)(&allPositions[0]), memSizePos))goto ERROR_PrepareCellMask;
  if(!memcpyHtoD(ptrMask, (void*)(&allMask[0]),      memSizeMask))goto ERROR_PrepareCellMask;

  __ALIGN(ptrPos);      if(!paramSetV( func, &offset, sizeof(ptrPos),  (void*)&ptrPos,  sizeof(ptrPos)))  goto ERROR_PrepareCellMask;
  __ALIGN(ptrMask);     if(!paramSetV( func, &offset, sizeof(ptrMask), (void*)&ptrMask, sizeof(ptrMask))) goto ERROR_PrepareCellMask;
  __ALIGN(stepAlpha);   if(!paramSetI( func, &offset, sizeof(int), stepAlpha))                            goto ERROR_PrepareCellMask;
  __ALIGN(BxG);         if(!paramSetI( func, &offset, sizeof(int), BxG))                                  goto ERROR_PrepareCellMask;
  __ALIGN(maskBase);    if(!paramSetUL(func, &offset, sizeof(ulong), maskBase, &maskBaseOffset))          goto ERROR_PrepareCellMask;
  __ALIGN(prj.wallX0);  if(!paramSetF( func, &offset, sizeof(float), prj.wallX0))                         goto ERROR_PrepareCellMask;
  __ALIGN(prj.wallX1);  if(!paramSetF( func, &offset, sizeof(float), prj.wallX1))                         goto ERROR_PrepareCellMask;
  __ALIGN(prj.wallY0);  if(!paramSetF( func, &offset, sizeof(float), prj.wallY0))                         goto ERROR_PrepareCellMask;
  __ALIGN(prj.wallY1);  if(!paramSetF( func, &offset, sizeof(float), prj.wallY1))                         goto ERROR_PrepareCellMask;
  __ALIGN(prj.wallZ0);  if(!paramSetF( func, &offset, sizeof(float), prj.wallZ0))                         goto ERROR_PrepareCellMask;
  __ALIGN(prj.wallX1);  if(!paramSetF( func, &offset, sizeof(float), prj.wallZ1))                         goto ERROR_PrepareCellMask;

  if(!paramSetSize(     func, offset))goto ERROR_PrepareCellMask;
  if(!funcSetBlockShape(func, maxBlockShape, 1, 1))goto ERROR_PrepareCellMask;

  if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellMask;
  if(!memcpyDtoH((void*)(&allMask[0]), ptrMask, memSizeMask))goto ERROR_PrepareCellMask;

  CellMaskSetLattice( &maskBase, 2);
  if(!memcpyHtoD(ptrPos,  (void*)(&allPositions[AxBxG]), memSizePos)) goto ERROR_PrepareCellMask;
  if(!memcpyHtoD(ptrMask, (void*)(&allMask[AxBxG]),      memSizeMask))goto ERROR_PrepareCellMask;
  __ALIGN(maskBase);
  if(!paramSetUL(func, &maskBaseOffset, 0, maskBase))goto ERROR_PrepareCellMask;
  if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellMask;
  if(!memcpyDtoH((void*)(&allMask[AxBxG]), ptrMask, memSizeMask))goto ERROR_PrepareCellMask;

  if(prj.matrixType == FCC) {
    CellMaskSetLattice( &maskBase, 3);
    if(!memcpyHtoD(ptrPos,  (void*)(&allPositions[2*AxBxG]), memSizePos)) goto ERROR_PrepareCellMask;
    if(!memcpyHtoD(ptrMask, (void*)(&allMask[2*AxBxG]),      memSizeMask))goto ERROR_PrepareCellMask;
    __ALIGN(maskBase);
    if(!paramSetUL(func, &maskBaseOffset, 0, maskBase))goto ERROR_PrepareCellMask;
    if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellMask;
    if(!memcpyDtoH((void*)(&allMask[2*AxBxG]), ptrMask, memSizeMask))goto ERROR_PrepareCellMask;

    CellMaskSetLattice( &maskBase, 4);
    if(!memcpyHtoD(ptrPos,  (void*)(&allPositions[3*AxBxG]), memSizePos)) goto ERROR_PrepareCellMask;
    if(!memcpyHtoD(ptrMask, (void*)(&allMask[3*AxBxG]),      memSizeMask))goto ERROR_PrepareCellMask;
    __ALIGN(maskBase);
    if(!paramSetUL(func, &maskBaseOffset, 0, maskBase))goto ERROR_PrepareCellMask;
    if(!launchGrid(func, maxUsedThred > maxThredPerBlock, maxUsedThred/maxThredPerBlock+1, 1))goto ERROR_PrepareCellMask;
    if(!memcpyDtoH((void*)(&allMask[3*AxBxG]), ptrMask, memSizeMask))goto ERROR_PrepareCellMask;
  }

  result = true;
  ERROR_PrepareCellMask:
  if(!memFree(ptrPos)) result = false;
  if(!memFree(ptrMask))result = false;
  return result;
}

bool CudaDevice::stabilisation(ProjectDataF * data, CalcDataF * calcData, int cntData, float stopLoopStress)
{
/*
extern "C" __global__ void StabilisationBase( float * stressOut, float3 * posOut, float3 * posIn,
                                              CellAttribute * attr, CellMate * mate, CalcDataF * data,
                                              int cellCount, int dataCount, int steps);
*/

  bool result = false, needLoop;
  QString kernel = "StabilisationBase";
  //QString kernel = "StabilisationLJ1";
  //QString kernel = "StabilisationLJ2";

  int offset = 0, offset2, offset3;
  int reg = REGS__StabilisationBase;

  int maxThredPerBlock = props.regsPerBlock / ((reg == 0) ? 1 : reg);
  int maxUsedThred = maxThredPerBlock;
  int maxBlockShape;
  int steps = (data->count / maxUsedThred) + 1;
  int l, i;
  size_t memSizeOut   = maxUsedThred  * sizeof(float);
  size_t memSizePos   = data->count   * sizeof(float3);
  size_t memSizeAttr  = data->count   * sizeof(CellAttribute);
  size_t memSizeMate  = data->count   * sizeof(CellMate);
  size_t memSizeData  = cntData       * sizeof(CalcDataF);

  CUfunction  func;
  CUdeviceptr ptrStressOut, ptrPosOut, ptrPos, ptrAttr, ptrMate, ptrData, tmpDevPtr;
  float * threadStress = 0;
  float curMax = 0.0f;
  if(maxThredPerBlock > props.maxThreadsPerBlock)maxThredPerBlock = props.maxThreadsPerBlock;
  if(maxUsedThred > maxThredPerBlock)maxBlockShape = maxThredPerBlock;
  else maxBlockShape = maxUsedThred;


  if(!moduleGetFunction(kernel.toUtf8().data(), &func))return result;

  if(!memAlloc(&ptrStressOut, memSizeOut))  return result;
  if(!memAlloc(&ptrPosOut,    memSizePos))  return result;
  if(!memAlloc(&ptrPos,       memSizePos))  return result;
  if(!memAlloc(&ptrAttr,      memSizeAttr)) return result;
  if(!memAlloc(&ptrMate,      memSizeMate)) return result;
  if(!memAlloc(&ptrData,      memSizeData)) return result;

  if(!memcpyHtoD(ptrPos,  (void*)(&data->pos[0]),   memSizePos)) goto ERROR_Stabilisation;
  if(!memcpyHtoD(ptrAttr, (void*)(&data->attr[0]),  memSizeAttr))goto ERROR_Stabilisation;
  if(!memcpyHtoD(ptrMate, (void*)(&data->mate[0]),  memSizeMate))goto ERROR_Stabilisation;
  if(!memcpyHtoD(ptrMate, (void*)(&calcData[0]),    memSizeData))goto ERROR_Stabilisation;

  __ALIGN(ptrStressOut);  if(!paramSetV( func, &offset, sizeof(ptrStressOut), (void*)&ptrStressOut, sizeof(ptrStressOut)))        goto ERROR_Stabilisation;
  __ALIGN(ptrPosOut);     if(!paramSetV( func, &offset, sizeof(ptrPosOut),    (void*)&ptrPosOut,    sizeof(ptrPosOut), &offset2)) goto ERROR_Stabilisation;
  __ALIGN(ptrPos);        if(!paramSetV( func, &offset, sizeof(ptrPos),       (void*)&ptrPos,       sizeof(ptrPos),    &offset3)) goto ERROR_Stabilisation;
  __ALIGN(ptrAttr);       if(!paramSetV( func, &offset, sizeof(ptrAttr),      (void*)&ptrAttr,      sizeof(ptrAttr)))             goto ERROR_Stabilisation;
  __ALIGN(ptrMate);       if(!paramSetV( func, &offset, sizeof(ptrMate),      (void*)&ptrMate,      sizeof(ptrMate)))             goto ERROR_Stabilisation;
  __ALIGN(ptrData);       if(!paramSetV( func, &offset, sizeof(ptrData),      (void*)&ptrData,      sizeof(ptrData)))             goto ERROR_Stabilisation;
  __ALIGN(data->count);   if(!paramSetI( func, &offset, sizeof(int), data->count))                                                goto ERROR_Stabilisation;
  __ALIGN(cntData);       if(!paramSetI( func, &offset, sizeof(int), cntData))                                                    goto ERROR_Stabilisation;
  __ALIGN(steps);         if(!paramSetI( func, &offset, sizeof(int), steps))                                                      goto ERROR_Stabilisation;

  if(!paramSetSize(     func, offset))             goto ERROR_Stabilisation;
  if(!funcSetBlockShape(func, maxBlockShape, 1, 1))goto ERROR_Stabilisation;
  if(!funcSetSharedSize(func, memSizeData))        goto ERROR_Stabilisation;

  threadStress = new float[maxUsedThred];

  for(l = 0 ; l < maxStabilLoops ; l++){
    if(!paramSetV(func, &offset2, 0, (void*)&ptrPosOut, sizeof(ptrPosOut)))goto ERROR_Stabilisation;
    if(!paramSetV(func, &offset3, 0, (void*)&ptrPos,    sizeof(ptrPos)))goto ERROR_Stabilisation;

    if(!launchGrid( func, maxUsedThred > maxThredPerBlock, maxUsedThred / maxThredPerBlock + 1, 1))goto ERROR_Stabilisation;

    if(!memcpyDtoH((void*)(threadStress), ptrStressOut, memSizeOut))goto ERROR_Stabilisation;

    needLoop = false;
    for(i = 0 ; i < maxUsedThred ; i++){
      if(threadStress[i] > stopLoopStress){
        needLoop = true;
        curMax = threadStress[i];
        break;
      }
    }
    if(!needLoop)break;

    qDebug() << "Stress: " << curMax;

    tmpDevPtr  = ptrPosOut;
    ptrPosOut  = ptrPos;
    ptrPos     = tmpDevPtr;

  }

  if(!memcpyDtoH((void*)(&data->pos[0]), ptrPosOut, memSizePos)) goto ERROR_Stabilisation;

  qDebug() << "Stabilisation loops: " << l;

  result = true;
  ERROR_Stabilisation:
  if(!memFree(ptrStressOut))result = false;
  if(!memFree(ptrPosOut))   result = false;
  if(!memFree(ptrPos))      result = false;
  if(!memFree(ptrAttr))     result = false;
  if(!memFree(ptrMate))     result = false;
  if(!memFree(ptrData))     result = false;

  return result;
}


bool CudaDevice::prepareMateOffset(CrystalMatrixType type, int matrix, int * mateID, int stepAlpha, int stepBeta, int stepGama){
  if(type == BCC){
    if(matrix == 1){
      mateID[PA_PB_PG] = stepAlpha*stepBeta*stepGama;
      mateID[MA_MB_MG] = stepAlpha*stepBeta*stepGama - stepBeta * stepGama - stepGama - 1;
      mateID[PA_PB_MG] = stepAlpha*stepBeta*stepGama - 1;
      mateID[MA_MB_PG] = stepAlpha*stepBeta*stepGama - stepBeta * stepGama - stepGama;
      mateID[PA_MB_PG] = stepAlpha*stepBeta*stepGama - stepGama;
      mateID[MA_PB_MG] = stepAlpha*stepBeta*stepGama - stepBeta * stepGama - 1;
      mateID[PA_MB_MG] = stepAlpha*stepBeta*stepGama - stepGama - 1;
      mateID[MA_PB_PG] = stepAlpha*stepBeta*stepGama - stepBeta * stepGama;
      for(int i=MA_PB_PG+1 ; i < MAX_MATES ; i++ )mateID[i]=0;
      return true;
    }
    if(matrix == 2){
      mateID[PA_PB_PG] = -stepAlpha*stepBeta*stepGama + stepBeta * stepGama + stepGama + 1;
      mateID[MA_MB_MG] = -stepAlpha*stepBeta*stepGama;
      mateID[PA_PB_MG] = -stepAlpha*stepBeta*stepGama + stepBeta * stepGama + stepGama;
      mateID[MA_MB_PG] = -stepAlpha*stepBeta*stepGama + 1;
      mateID[PA_MB_PG] = -stepAlpha*stepBeta*stepGama + stepBeta * stepGama + 1;
      mateID[MA_PB_MG] = -stepAlpha*stepBeta*stepGama + stepGama;
      mateID[PA_MB_MG] = -stepAlpha*stepBeta*stepGama + stepBeta * stepGama;
      mateID[MA_PB_PG] = -stepAlpha*stepBeta*stepGama + stepGama + 1;
      for(int i=MA_PB_PG+1 ; i < MAX_MATES ; i++ )mateID[i]=0;
      return true;
    }
  }

  if(type == FCC){
    if(matrix == 1){
      mateID[PA_PB] =      stepAlpha*stepBeta*stepGama;
      mateID[MA_MB] =      stepAlpha*stepBeta*stepGama - stepBeta * stepGama - stepGama;
      mateID[PA_MB] =      stepAlpha*stepBeta*stepGama - stepGama;
      mateID[MA_PB] =      stepAlpha*stepBeta*stepGama - stepBeta * stepGama;
      mateID[PA_PG] =  2 * stepAlpha*stepBeta*stepGama;
      mateID[MA_MG] =  2 * stepAlpha*stepBeta*stepGama - stepBeta * stepGama - 1;
      mateID[PA_MG] =  2 * stepAlpha*stepBeta*stepGama - 1;
      mateID[MA_PG] =  2 * stepAlpha*stepBeta*stepGama - stepBeta * stepGama;
      mateID[PB_PG] =  3 * stepAlpha*stepBeta*stepGama;
      mateID[MB_MG] =  3 * stepAlpha*stepBeta*stepGama - stepGama - 1;
      mateID[PB_MG] =  3 * stepAlpha*stepBeta*stepGama - 1;
      mateID[MB_PG] =  3 * stepAlpha*stepBeta*stepGama - stepGama;
      for(int i=MB_PG+1 ; i < MAX_MATES ; i++ )mateID[i]=0;
      return true;
    }

    if(matrix == 2){
      mateID[PA_PB] =     -stepAlpha*stepBeta*stepGama + stepBeta * stepGama + stepGama;
      mateID[MA_MB] =     -stepAlpha*stepBeta*stepGama;
      mateID[PA_MB] =     -stepAlpha*stepBeta*stepGama + stepBeta * stepGama;
      mateID[MA_PB] =     -stepAlpha*stepBeta*stepGama + stepGama;
      mateID[PA_PG] =  2 * stepAlpha*stepBeta*stepGama + stepBeta * stepGama;
      mateID[MA_MG] =  2 * stepAlpha*stepBeta*stepGama - 1;
      mateID[PA_MG] =  2 * stepAlpha*stepBeta*stepGama + stepBeta * stepGama - 1;
      mateID[MA_PG] =  2 * stepAlpha*stepBeta*stepGama;
      mateID[PB_PG] =      stepAlpha*stepBeta*stepGama + stepGama;
      mateID[MB_MG] =      stepAlpha*stepBeta*stepGama - 1;
      mateID[PB_MG] =      stepAlpha*stepBeta*stepGama + stepGama - 1;
      mateID[MB_PG] =      stepAlpha*stepBeta*stepGama;
      for(int i=MB_PG+1 ; i < MAX_MATES ; i++ )mateID[i]=0;
      return true;
    }

    if(matrix == 3){
      mateID[PA_PB] =      stepAlpha*stepBeta*stepGama + stepBeta * stepGama;
      mateID[MA_MB] =      stepAlpha*stepBeta*stepGama - stepGama;
      mateID[PA_MB] =      stepAlpha*stepBeta*stepGama + stepBeta * stepGama - stepGama;
      mateID[MA_PB] =      stepAlpha*stepBeta*stepGama;
      mateID[PA_PG] = -2 * stepAlpha*stepBeta*stepGama + stepBeta * stepGama + 1;
      mateID[MA_MG] = -2 * stepAlpha*stepBeta*stepGama;
      mateID[PA_MG] = -2 * stepAlpha*stepBeta*stepGama + stepBeta * stepGama;
      mateID[MA_PG] = -2 * stepAlpha*stepBeta*stepGama + 1;
      mateID[PB_PG] =     -stepAlpha*stepBeta*stepGama + 1;
      mateID[MB_MG] =     -stepAlpha*stepBeta*stepGama - stepGama;
      mateID[PB_MG] =     -stepAlpha*stepBeta*stepGama;
      mateID[MB_PG] =     -stepAlpha*stepBeta*stepGama - stepGama + 1;
      for(int i=MB_PG+1 ; i < MAX_MATES ; i++ )mateID[i]=0;
      return true;
    }

    if(matrix == 4){
      mateID[PA_PB] =     -stepAlpha*stepBeta*stepGama + stepGama;
      mateID[MA_MB] =     -stepAlpha*stepBeta*stepGama - stepBeta * stepGama;
      mateID[PA_MB] =     -stepAlpha*stepBeta*stepGama;
      mateID[MA_PB] =     -stepAlpha*stepBeta*stepGama - stepBeta * stepGama + stepGama;
      mateID[PA_PG] = -2 * stepAlpha*stepBeta*stepGama + 1;
      mateID[MA_MG] = -2 * stepAlpha*stepBeta*stepGama - stepBeta * stepGama;
      mateID[PA_MG] = -2 * stepAlpha*stepBeta*stepGama;
      mateID[MA_PG] = -2 * stepAlpha*stepBeta*stepGama - stepBeta * stepGama + 1;
      mateID[PB_PG] = -3 * stepAlpha*stepBeta*stepGama + stepGama + 1;
      mateID[MB_MG] = -3 * stepAlpha*stepBeta*stepGama;
      mateID[PB_MG] = -3 * stepAlpha*stepBeta*stepGama + stepGama;
      mateID[MB_PG] = -3 * stepAlpha*stepBeta*stepGama + 1;
      for(int i=MB_PG+1 ; i < MAX_MATES ; i++ )mateID[i]=0;
      return true;
    }
  }
  return false;
}


CudaDevice * CudaDevice::createCudaDevice(int devId)
{
  CUresult  result;
  int count;

  if((result = cuInit(0)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("Error: %1 in method cuInit").arg(errorString(result));
    return 0;
  }

  if((result = cuDeviceGetCount(&count)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("Error: %1 in method cuDeviceGetCount").arg(errorString(result));
    return 0;
  }
  int idUsed = devId;
  if(idUsed >= count){
    idUsed = count - 1;
    qDebug() << QObject::trUtf8("Warning: Invalid CUDA device ID, using %1 !").arg(idUsed);
  } else if(idUsed < 0){
    idUsed = 0;
    qDebug() << QObject::trUtf8("Warning: Invalid CUDA device ID, using %1 !").arg(idUsed);
  }
  CudaDevice * out = new CudaDevice(idUsed);
  if(!out->initialized() || !out->createContext()){
    delete out;
    return 0;
  }
  return out;
}

void CudaDevice::printCudaDevices()
{
  CUresult  result;
  int count;

  if((result = cuInit(0)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("Error: %1 in method cuInit").arg(errorString(result));
    return;
  }

  if((result = cuDeviceGetCount(&count)) != CUDA_SUCCESS){
    qDebug() << QObject::trUtf8("Error: %1 in method cuDeviceGetCount").arg(errorString(result));
    return;
  }

  for(int i = 0 ; i < count ; i++){
    CudaDevice * dev = new CudaDevice(i);
    if(!dev->initialized()){
      delete dev;
      return;
    }
    dev->createContext();
    qDebug() << qPrintable(dev->ToString());
    dev->detachContext();
    delete dev;
  }
}


QString CudaDevice::ToString(bool toHtmlString){
  QString text, s1, s2, s3;
  if(toHtmlString){
    s1 = "<tr><td><b>";
    s2 = "</b></td><td>";
    s3 = "</td></tr>";
    text.append("<table border=\"1\" cellpadding=\"5\" >");
  }else{
    s1 = "";
    s2 = ": ";
    s3 = "\n";
  }

  text.append(QObject::trUtf8("%1ID%2%4/%5%3")                                    .arg(s1).arg(s2).arg(s3).arg(_id).arg(device)                                                   );
  text.append(QObject::trUtf8("%1Name%2%4%3")                                     .arg(s1).arg(s2).arg(s3).arg(_name)                                                             );
  text.append(QObject::trUtf8("%1Ready%2%4%3")                                    .arg(s1).arg(s2).arg(s3).arg(isInit == true ? QString("Yes") : QString("No")));
  text.append(QObject::trUtf8("%1Compute capability%2%4.%5%3")                    .arg(s1).arg(s2).arg(s3).arg(_cMaj).arg(_cMin)                                                   );
  text.append(QObject::trUtf8("%1Double precission%2%4%3")                        .arg(s1).arg(s2).arg(s3).arg(isDoublePrecission() == true ? QString("Yes") : QString("No"))    );
  text.append(QObject::trUtf8("%1Total memory%2%4%3")                             .arg(s1).arg(s2).arg(s3).arg(_totalMem)                                                         );
  text.append(QObject::trUtf8("%1Free memory%2%4%3")                              .arg(s1).arg(s2).arg(s3).arg(_infoMemFree)                                                         );
  text.append(QObject::trUtf8("%1Max number of threads per block%2%4%3")          .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK])             );
  text.append(QObject::trUtf8("%1Max block dimension X%2%4%3")                    .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_X])                   );
  text.append(QObject::trUtf8("%1Max block dimension Y%2%4%3")                    .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_Y])                   );
  text.append(QObject::trUtf8("%1Max block dimension Z%2%4%3")                    .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_Z])                   );
  text.append(QObject::trUtf8("%1Max grid dimension X%2%4%3")                     .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_X])                    );
  text.append(QObject::trUtf8("%1Max grid dimension Y%2%4%3")                     .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_Y])                    );
  text.append(QObject::trUtf8("%1Max grid dimension Z%2%4%3")                     .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_Z])                    );
  text.append(QObject::trUtf8("%1Max shared memory available per block%2%4%3")    .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_BLOCK])       );
  text.append(QObject::trUtf8("%1Memory available on device%2%4%3")               .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_TOTAL_CONSTANT_MEMORY])             );
  text.append(QObject::trUtf8("%1Warp size in threads%2%4%3")                     .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_WARP_SIZE])                         );
  text.append(QObject::trUtf8("%1Max pitch allowed by memory copies%2%4%3")       .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAX_PITCH])                         );
  text.append(QObject::trUtf8("%1Max number of 32-bit registers per block%2%4%3") .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAX_REGISTERS_PER_BLOCK])           );
  text.append(QObject::trUtf8("%1Peak clock frequency in kilohertz%2%4%3")        .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_CLOCK_RATE])                        );
  text.append(QObject::trUtf8("%1Alignment requirement for textures%2%4%3")       .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_TEXTURE_ALIGNMENT])                 );
  text.append(QObject::trUtf8("%1Device can copy / execute concurrently%2%4%3")   .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_GPU_OVERLAP])                       );
  text.append(QObject::trUtf8("%1Number of multiprocessors on device%2%4%3")      .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT])              );
  text.append(QObject::trUtf8("%1Run time limit on kernels%2%4%3")                .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_KERNEL_EXEC_TIMEOUT])               );
  text.append(QObject::trUtf8("%1Is integrated with host memory%2%4%3")           .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_INTEGRATED])                        );
  text.append(QObject::trUtf8("%1Map host memory into CUDA%2%4%3")                .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_CAN_MAP_HOST_MEMORY])               );
  text.append(QObject::trUtf8("%1Compute mode%2%4%3")                             .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_COMPUTE_MODE])                      );
  text.append(QObject::trUtf8("%1Max 1D texture width%2%4%3")                     .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE1D_WIDTH])           );
  text.append(QObject::trUtf8("%1Max 2D texture width%2%4%3")                     .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_WIDTH])           );
  text.append(QObject::trUtf8("%1Max 2D texture height%2%4%3")                    .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_HEIGHT])          );
  text.append(QObject::trUtf8("%1Max 3D texture width%2%4%3")                     .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_WIDTH])           );
  text.append(QObject::trUtf8("%1Max 3D texture height%2%4%3")                    .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_HEIGHT])          );
  text.append(QObject::trUtf8("%1Max 3D texture depth%2%4%3")                     .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_DEPTH])           );
  text.append(QObject::trUtf8("%1Max texture array width%2%4%3")                  .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_ARRAY_WIDTH])     );
  text.append(QObject::trUtf8("%1Max texture array height%2%4%3")                 .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_ARRAY_HEIGHT])    );
  text.append(QObject::trUtf8("%1Max slices in a texture array%2%4%3")            .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_ARRAY_NUMSLICES]) );
  text.append(QObject::trUtf8("%1Alignment requirement for surfaces%2%4%3")       .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_SURFACE_ALIGNMENT])                 );
  text.append(QObject::trUtf8("%1Can execute multiple kernels concurrently%2%4%3").arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_CONCURRENT_KERNELS])                );
  text.append(QObject::trUtf8("%1Has ECC support enabled%2%4%3")                  .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_ECC_ENABLED])                       );
  text.append(QObject::trUtf8("%1PCI bus ID of the device%2%4%3")                 .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_PCI_BUS_ID])                        );
  text.append(QObject::trUtf8("%1PCI device ID of the device%2%4%3")              .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_PCI_DEVICE_ID])                     );
  text.append(QObject::trUtf8("%1Device is using TCC driver model%2%4%3")         .arg(s1).arg(s2).arg(s3).arg(attribute[CU_DEVICE_ATTRIBUTE_TCC_DRIVER])                        );
  if(toHtmlString)text.append("</table>");
  return text;
}

QString CudaDevice::errorString(CUresult val){
  switch(val){
  case(CUDA_SUCCESS):                               return QObject::trUtf8("[%1]: Success").arg(val);
  case(CUDA_ERROR_INVALID_VALUE):                   return QObject::trUtf8("[%1]: Invalid value").arg(val);
  case(CUDA_ERROR_OUT_OF_MEMORY):                   return QObject::trUtf8("[%1]: Out of memory").arg(val);
  case(CUDA_ERROR_NOT_INITIALIZED):                 return QObject::trUtf8("[%1]: Driver not initialized").arg(val);
  case(CUDA_ERROR_DEINITIALIZED):                   return QObject::trUtf8("[%1]: Driver deinitialized").arg(val);
  case(CUDA_ERROR_NO_DEVICE):                       return QObject::trUtf8("[%1]: No CUDA-capable device available").arg(val);
  case(CUDA_ERROR_INVALID_DEVICE):                  return QObject::trUtf8("[%1]: Invalid device").arg(val);
  case(CUDA_ERROR_INVALID_IMAGE):                   return QObject::trUtf8("[%1]: Invalid kernel image").arg(val);
  case(CUDA_ERROR_INVALID_CONTEXT):                 return QObject::trUtf8("[%1]: Invalid context").arg(val);
  case(CUDA_ERROR_CONTEXT_ALREADY_CURRENT):         return QObject::trUtf8("[%1]: Context already current").arg(val);
  case(CUDA_ERROR_MAP_FAILED):                      return QObject::trUtf8("[%1]: Map failed").arg(val);
  case(CUDA_ERROR_UNMAP_FAILED):                    return QObject::trUtf8("[%1]: Unmap failed").arg(val);
  case(CUDA_ERROR_ARRAY_IS_MAPPED):                 return QObject::trUtf8("[%1]: Array is mapped").arg(val);
  case(CUDA_ERROR_ALREADY_MAPPED):                  return QObject::trUtf8("[%1]: Already mapped").arg(val);
  case(CUDA_ERROR_NO_BINARY_FOR_GPU):               return QObject::trUtf8("[%1]: No binary for GPU").arg(val);
  case(CUDA_ERROR_ALREADY_ACQUIRED):                return QObject::trUtf8("[%1]: Already acquired").arg(val);
  case(CUDA_ERROR_NOT_MAPPED):                      return QObject::trUtf8("[%1]: Not mapped").arg(val);
  case(CUDA_ERROR_NOT_MAPPED_AS_ARRAY):             return QObject::trUtf8("[%1]: Not mapped as array").arg(val);
  case(CUDA_ERROR_NOT_MAPPED_AS_POINTER):           return QObject::trUtf8("[%1]: Not mapped as poiner").arg(val);
  case(CUDA_ERROR_ECC_UNCORRECTABLE):               return QObject::trUtf8("[%1]: ECC uncorrectable").arg(val);
  case(CUDA_ERROR_UNSUPPORTED_LIMIT):               return QObject::trUtf8("[%1]: Unsuported limit").arg(val);
  case(CUDA_ERROR_INVALID_SOURCE):                  return QObject::trUtf8("[%1]: Invalid source").arg(val);
  case(CUDA_ERROR_FILE_NOT_FOUND):                  return QObject::trUtf8("[%1]: File not found").arg(val);
  case(CUDA_ERROR_SHARED_OBJECT_SYMBOL_NOT_FOUND):  return QObject::trUtf8("[%1]: Shared object symbol not found").arg(val);
  case(CUDA_ERROR_SHARED_OBJECT_INIT_FAILED):       return QObject::trUtf8("[%1]: Shared object init failed").arg(val);
  case(CUDA_ERROR_OPERATING_SYSTEM):                return QObject::trUtf8("[%1]: Operating system").arg(val);
  case(CUDA_ERROR_INVALID_HANDLE):                  return QObject::trUtf8("[%1]: Invalid handle").arg(val);
  case(CUDA_ERROR_NOT_FOUND):                       return QObject::trUtf8("[%1]: Not found").arg(val);
  case(CUDA_ERROR_NOT_READY):                       return QObject::trUtf8("[%1]: CUDA not ready").arg(val);
  case(CUDA_ERROR_LAUNCH_FAILED):                   return QObject::trUtf8("[%1]: Launch failed").arg(val);
  case(CUDA_ERROR_LAUNCH_OUT_OF_RESOURCES):         return QObject::trUtf8("[%1]: Launch exceeded resources").arg(val);
  case(CUDA_ERROR_LAUNCH_TIMEOUT):                  return QObject::trUtf8("[%1]: Launch exceeded timeout").arg(val);
  case(CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING):   return QObject::trUtf8("[%1]: Launch with incompatible texturing").arg(val);
  case(CUDA_ERROR_UNKNOWN):                         return QObject::trUtf8("[%1]: Unknown error");
  default:                                          return QObject::trUtf8("[%1]: Undefined error").arg(val);
  }
}

QString CudaDevice::attributeString(CUdevice_attribute val){
  switch(val){
  case(CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK):              return QObject::trUtf8("[%1]: Maximum number of threads per block").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_X):                    return QObject::trUtf8("[%1]: Maximum block dimension X").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_Y):                    return QObject::trUtf8("[%1]: Maximum block dimension Y").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_Z):                    return QObject::trUtf8("[%1]: Maximum block dimension Z").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_X):                     return QObject::trUtf8("[%1]: Maximum grid dimension X").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_Y):                     return QObject::trUtf8("[%1]: Maximum grid dimension Y").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_Z):                     return QObject::trUtf8("[%1]: Maximum grid dimension Z").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_BLOCK):        return QObject::trUtf8("[%1]: Maximum shared memory available per block in bytes").arg(val);
  case(CU_DEVICE_ATTRIBUTE_TOTAL_CONSTANT_MEMORY):              return QObject::trUtf8("[%1]: Memory available on device for __constant__ variables in a CUDA C kernel in bytes").arg(val);
  case(CU_DEVICE_ATTRIBUTE_WARP_SIZE):                          return QObject::trUtf8("[%1]: Warp size in threads").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAX_PITCH):                          return QObject::trUtf8("[%1]: Maximum pitch in bytes allowed by memory copies").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAX_REGISTERS_PER_BLOCK):            return QObject::trUtf8("[%1]: Maximum number of 32-bit registers available per block").arg(val);
  case(CU_DEVICE_ATTRIBUTE_CLOCK_RATE):                         return QObject::trUtf8("[%1]: Peak clock frequency in kilohertz").arg(val);
  case(CU_DEVICE_ATTRIBUTE_TEXTURE_ALIGNMENT):                  return QObject::trUtf8("[%1]: Alignment requirement for textures").arg(val);
  case(CU_DEVICE_ATTRIBUTE_GPU_OVERLAP):                        return QObject::trUtf8("[%1]: Device can possibly copy memory and execute a kernel concurrently").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT):               return QObject::trUtf8("[%1]: Number of multiprocessors on device").arg(val);
  case(CU_DEVICE_ATTRIBUTE_KERNEL_EXEC_TIMEOUT):                return QObject::trUtf8("[%1]: Specifies whether there is a run time limit on kernels").arg(val);
  case(CU_DEVICE_ATTRIBUTE_INTEGRATED):                         return QObject::trUtf8("[%1]: Device is integrated with host memory").arg(val);
  case(CU_DEVICE_ATTRIBUTE_CAN_MAP_HOST_MEMORY):                return QObject::trUtf8("[%1]: Device can map host memory into CUDA address space").arg(val);
  case(CU_DEVICE_ATTRIBUTE_COMPUTE_MODE):                       return QObject::trUtf8("[%1]: Compute mode").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE1D_WIDTH):            return QObject::trUtf8("[%1]: Maximum 1D texture width").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_WIDTH):            return QObject::trUtf8("[%1]: Maximum 2D texture width").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_HEIGHT):           return QObject::trUtf8("[%1]: Maximum 2D texture height").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_WIDTH):            return QObject::trUtf8("[%1]: Maximum 3D texture width").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_HEIGHT):           return QObject::trUtf8("[%1]: Maximum 3D texture height").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE3D_DEPTH):            return QObject::trUtf8("[%1]: Maximum 3D texture depth").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_ARRAY_WIDTH):      return QObject::trUtf8("[%1]: Maximum texture array width").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_ARRAY_HEIGHT):     return QObject::trUtf8("[%1]: Maximum texture array height").arg(val);
  case(CU_DEVICE_ATTRIBUTE_MAXIMUM_TEXTURE2D_ARRAY_NUMSLICES):  return QObject::trUtf8("[%1]: Maximum slices in a texture array").arg(val);
  case(CU_DEVICE_ATTRIBUTE_SURFACE_ALIGNMENT):                  return QObject::trUtf8("[%1]: Alignment requirement for surfaces").arg(val);
  case(CU_DEVICE_ATTRIBUTE_CONCURRENT_KERNELS):                 return QObject::trUtf8("[%1]: Device can possibly execute multiple kernels concurrently").arg(val);
  case(CU_DEVICE_ATTRIBUTE_ECC_ENABLED):                        return QObject::trUtf8("[%1]: Device has ECC support enabled").arg(val);
  case(CU_DEVICE_ATTRIBUTE_PCI_BUS_ID):                         return QObject::trUtf8("[%1]: PCI bus ID of the device").arg(val);
  case(CU_DEVICE_ATTRIBUTE_PCI_DEVICE_ID):                      return QObject::trUtf8("[%1]: PCI device ID of the device").arg(val);
  case(CU_DEVICE_ATTRIBUTE_TCC_DRIVER):                         return QObject::trUtf8("[%1]: Device is using TCC driver model").arg(val);
  default:                                                      return QObject::trUtf8("[%1]: Undefined attribute type").arg(val);
  }
}

QString CudaDevice::attributeString(int val){
  return attributeString((CUdevice_attribute)val);
}
